@@ -10,7 +10,7 @@ License: MIT (see LICENSE for details)
 """
 
 __author__ = 'Naphtha Nepanthez'
-__version__ = '0.0.8'
+__version__ = '0.2.2'
 __license__ = 'MIT' # SEE LICENSE FILE
 __all__ = [
     'Interface',
@@ -20,9 +20,9 @@ __all__ = [
 ]
 
 import os, struct, xxhash, pickle, hashlib, io, math, threading, functools
-import lz4.frame
+import zlib, time, random, tempfile
 
-from filelock import Timeout, FileLock, SoftFileLock
+from filelock import FileLock
 from collections.abc import MutableMapping
 
 class Exceptions(object):
@@ -44,38 +44,153 @@ class Exceptions(object):
     class InvalidFormat(Exception):
         pass
 
+class RRCCache(object):
+    def __init__(self, size = 8192):
+        self.size = size
+        self.foward_mapping = {}
+        self.backward_mapping = {}
+        
+    def set(self, key, value):
+        self.foward_mapping[key] = value
+        self.backward_mapping[value] = key
+        
+    def has(self, key):
+        return (key in self.foward_mapping)
+        
+    def keys(self):
+        return self.foward_mapping.keys()
+        
+    def get(self, key):
+        return self.foward_mapping[key]
+        
+    def get_key(self, value):
+        return self.backward_mapping[value]
+        
+    def remove(self, key):
+        del self.backward_mapping[self.foward_mapping[key]]
+        del self.foward_mapping[key]
+        
+    def random_key(self):
+        return random.choice(self.backward_mapping)
+        
+    def check_full(self):
+        if len(self.foward_mapping) > self.size:
+            self.push_out()
+            
+    def push_out(self):
+        self.remove(self.random_key())
+        
+    def length(self):
+        return len(self.foward_mapping)
+
+class ArbRemovalCache(RRCCache):
+    def push_out(self):
+        # Only works on very latest versions of Python
+        self.foward_mapping.popitem()
+        self.backward_mapping.popitem()
+
+class OrderedRemovalCache(RRCCache):
+    def push_out(self):
+        self.remove(list(self.keys())[0])
+
+class CacheDict(MutableMapping):
+    def __init__(self, cache, *args, **kwargs):
+        self.mapping = cache(*args, **kwargs)
+        
+    def __getitem__(self, key):
+        return self.mapping.get(key)
+        
+    def __delitem__(self, key):
+        self.mapping.remove(key)
+        
+    def __setitem__(self, key, value):
+        self.mapping.set(key, value)
+        
+    def __iter__(self):
+        return iter(self.mapping.keys())
+            
+    def __len__(self):
+        return self.mapping.length()
+
 class AtomicCachingFileLock(FileLock):
     def reset_cache(self):
         self.cache = {}
         self.cache['cells'] = {}
+        self.cache['keypos'] = RRCCache()
+        self.cache['blocks'] = CacheDict(RRCCache)
         self.cache['index_cell_translation'] = {}
         self.cache['safe_indexes'] = set()
     
+    def _cap32(self, i):
+        return i & ((2 ** 32) - 1)
+    
+    def _session(self):
+        return struct.pack('<III',
+            self._cap32(threading.get_ident()),
+            self._cap32(os.getpid()),
+            self._cap32(round(time.time() / 3600))
+        ) + self.id
+    
+    def _write_session(self):
+        if os.path.getsize(self._original_path) != 0:
+            self.handle.seek(0)
+            self.handle.write(self._session())
+            
+    def _same_session(self):
+        if os.path.getsize(self._original_path) != 0:
+            self.handle.seek(0)
+            cses = self._session()
+            if self.handle.read(len(cses)) == cses:
+                return True
+            else:
+                return False
+        else:
+            return False
+    
+    def _templk(self, i):
+        return os.path.join(
+            tempfile.gettempdir(), 
+            xxhash.xxh64(i).hexdigest() + '.ylf'
+        )
+    
     def __init__(self, *args, **kwargs):
-        self._original_path = args[0]
         args = list(args)
-        args[0] += '.lock'
-        self._super_thread_lock = threading.Lock()
+        
+        self.id = os.urandom(4)
+        
+        self._original_path = args[0]
+        
+        self.lockfile = self._templk(self._original_path + '.lock')
+        
+        args[0] = self.lockfile
         super().__init__(*args, **kwargs)
+        
         self.reset_cache()
+        
         self.handle = None
         
+    def acquire(self, *args, **kwargs):
+        super().acquire(*args, **kwargs)
+        
     def _acquire(self, *args, **kwargs):
-        self._super_thread_lock.acquire()
         super()._acquire(*args, **kwargs)
-        self.reset_cache()
+        
         self.handle = open(self._original_path, 'rb+')
         
+        if not self._same_session():
+            self.reset_cache()
+            self._write_session()
+        
     def _release(self, *args, **kwargs):
-        self._super_thread_lock.release()
         super()._release(*args, **kwargs)
-        self.reset_cache()
+        
         self.handle.close()
 
 class Interface(object):
-    _index_header_pattern = '<?IHH'  # 9  bytes
-    _index_cell_pattern   = '<?QIHQ' # 22 bytes
-    _identity_header      = b'YUN'   # 3 bytes
+    _index_header_pattern = '<?xxxIHH'  # 12 bytes
+    _index_cell_pattern   = '<?xQIHQ'   # 24 bytes
+    _yunyun_header        = b'YUNYUN01' # 8  bytes
+    _identity_header      = _yunyun_header.rjust(len(_yunyun_header)+16, b'#')
     _cache_size           = 4096
     
     def __init__(self, path, index_size = 4096, block_size = 4096):
@@ -103,7 +218,8 @@ class Interface(object):
             new = True
         else:
             hdr = open(self.path, 'rb').read(len(self._identity_header))
-            if hdr != self._identity_header:
+            cutlen = len(_yunyun_header)
+            if hdr[:-cutlen] != self._identity_header[:-cutlen]:
                 raise Exceptions.InvalidFormat(
                     'Not a YunYun file!'
                 )
@@ -152,14 +268,12 @@ class Interface(object):
             data
         )
         
-    @functools.lru_cache(maxsize=_cache_size)
     def readIndexHeader(self, index: bytes):
         return struct.unpack(
             self._index_header_pattern,
             index[len(self._identity_header):]
         )
     
-    @functools.lru_cache(maxsize=_cache_size)
     def readIndexCell(self, cell: bytes):
         return struct.unpack(
             self._index_cell_pattern,
@@ -206,6 +320,10 @@ class Interface(object):
         ]
     
     def getIndexesCells(self):
+        # TODO: This function is absolutely awful. I did my best to optimize
+        # it, but it's still an extremely intensive function to run. It's just
+        # super difficult.
+        
         with self.lock:
             indexes = self.getIndexes()
                 
@@ -216,27 +334,31 @@ class Interface(object):
                 
                 index_pointer = x[0] + self._index_headersize
                 f.seek(index_pointer)
-                    
-                allcells = io.BytesIO(f.read(
-                    self._indexes * self._index_cellsize
-                ))
                 
-                for z in self.generateIndexPositions(index_pointer):
-                    if z not in self.lock.cache['cells']:
-                        read = allcells.read(self._index_cellsize)
-                        self.lock.cache['cells'][z] = (
-                            self.readIndexCell(read)
-                        )
-                        self.lock.cache['index_cell_translation'][z] = x[0]
-                    else:
-                        allcells.seek(self._index_cellsize, 1)
-                        
+                positions = self.generateIndexPositions(index_pointer)
+                
+                cells = map(lambda z: (
+                    z,
+                    self.readIndexCell(f.read(self._index_cellsize))
+                ), positions)
+                self.lock.cache['cells'].update(cells)
+                
+                translations = map(lambda z: (z, x[0]), positions)
+                self.lock.cache['index_cell_translation'].update(translations)   
+                
                 self.lock.cache['safe_indexes'].add(x[0])
                     
         return self.lock.cache['cells']
     
     def markCellModified(self, pos):
         with self.lock:
+            try:
+                self.lock.cache['keypos'].remove(
+                    self.lock.cache['keypos'].get_key(pos)
+                )
+            except KeyError:
+                pass
+            
             try:
                 del self.lock.cache['cells'][pos]
             except KeyError:
@@ -278,22 +400,34 @@ class Interface(object):
               
     def keyExists(self, key: bytes):
         with self.lock:
+            if self.lock.cache['keypos'].has(key):
+                return self.lock.cache['keypos'].get(key)
+            
             hkey = xxhash.xxh64(key).intdigest()
-            for k, v in self.getIndexesCells().items():
-                if (v[1] == hkey
-                    and v[0] == True):
-                        
-                    return k
+            
+            vlt = 0
+            
+            # Pypy support
+            try:
+                cells = reversed(self.getIndexesCells().items())
+            except TypeError:
+                cells = self.getIndexesCells().items()
+            
+            for k, v in cells:
+                if (v[1] == hkey and v[0] == True):
+                    vlt = k
                     
-            return 0
+                    break
+            
+            self.lock.cache['keypos'].set(key, vlt)
+            
+            return vlt
                
     def requestFreeIndexCell(self):
         with self.lock:
-            ret = None
             while True:
                 for k, v in self.getIndexesCells().items():
                     if (v[0] == False):
-                            
                         return k
                         
                 self.createIndex()
@@ -312,17 +446,22 @@ class Interface(object):
             key_exists = self.keyExists(key)
             if not key_exists:
                 key_exists = self.requestFreeIndexCell()
+                try:
+                    self.lock.cache['keypos'].remove(key)
+                except KeyError:
+                    pass
                 
                 f.seek(key_exists)
                 cell = self.readIndexCell(f.read(self._index_cellsize))
                 
+                blank = b'\x00' * self._block_size
                 if cell[2] == 0:
                     f.seek(0, 2)
                     location = f.tell()
                     if hard:
-                        f.write(b'\x00' * self._block_size)
+                        f.write(blank)
                     else:
-                        f.truncate(location + self._block_size)
+                        f.truncate(location + len(blank))
                 else:
                     location = cell[2]
                 
@@ -336,6 +475,7 @@ class Interface(object):
                 ))
                 
                 self.markCellModified(key_exists)
+                self.setBlockCacheNewValue(key, blank)
                 
             valhash = xxhash.xxh64(value).intdigest()
 
@@ -355,6 +495,7 @@ class Interface(object):
             ))
             
             self.markCellModified(key_exists)
+            self.setBlockCacheNewValue(key, value)
             
             f.seek(cell[2])
             f.write(value)
@@ -377,6 +518,7 @@ class Interface(object):
                 ))
                 
                 self.markCellModified(key_exists)
+                self.invalidateBlockCacheKey(key)
             else:
                 raise Exceptions.BlockNotFound('!DELT Key: {0}'.format(
                     key.hex()
@@ -406,13 +548,27 @@ class Interface(object):
                 ))
                 
                 self.markCellModified(key_exists)
+                self.invalidateBlockCacheKey(key)
             else:
                 raise Exceptions.BlockNotFound('!RENM Key: {0}'.format(
                     key.hex()
                 ))
+              
+    def invalidateBlockCacheKey(self, key):
+        with self.lock:
+            try:
+                del self.lock.cache['blocks'][key]
+            except KeyError:
+                pass
+            
+    def setBlockCacheNewValue(self, key, value):
+        self.lock.cache['blocks'][key] = value
                     
     def readBlock(self, key: bytes):
         with self.lock:
+            if key in self.lock.cache['blocks']:
+                return self.lock.cache['blocks'][key]
+            
             f = self.lock.handle
             key_exists = self.keyExists(key)
             if key_exists:
@@ -420,7 +576,8 @@ class Interface(object):
                 cell = self.readIndexCell(f.read(self._index_cellsize))
                 
                 f.seek(cell[2])
-                return f.read(cell[3])
+                self.setBlockCacheNewValue(key, f.read(cell[3]))
+                return self.lock.cache['blocks'][key]
             else:
                 raise Exceptions.BlockNotFound('!READ Key: {0}'.format(
                     key.hex()
@@ -494,10 +651,20 @@ class MultiblockHandler(Interface):
                 ))
         
     def _getNodeProperties(self, key: bytes) -> dict:
-        return pickle.loads(self.readBlock(key))
+        op = self.readBlock(key)
+        blocks, size = struct.unpack('<II', op[:8])
+        
+        return {
+            'key': op[8:],
+            'blocks': blocks,
+            'size': size
+        }
     
     def _setNodeProperties(self, key: bytes, properties: dict):
-        self.writeBlock(key, pickle.dumps(properties))
+        self.writeBlock(key, struct.pack('<II',
+            properties['blocks'],
+            properties['size']
+        ) + properties['key'])
 
     class MultiblockFileHandle(object):
         def __init__(self, interface, key: bytes):
@@ -675,100 +842,121 @@ class MultiblockHandler(Interface):
 
 class Shelve(MutableMapping):
     def __init__(self, *args, **kwargs):
-        self.mapping = MultiblockHandler(*args, **kwargs)
+        self._lock = threading.Lock()
+        self.trackkeys = True
         
-        if self.mapping._block_size < 96:
-            raise Exceptions.InvalidFormat(
-                'Shelve mapping block size must be at least 96 bytes.'
-            )
-        
-        self._key_node_name = self._hash_key(b'__KEYS__')
-        
-        with self.mapping.lock:
-            first = False
-            if not self.mapping.nodeExists(self._key_node_name):
-                self.mapping.makeNode(self._key_node_name)
-                first = True
+        with self._lock:
+            self.mapping = MultiblockHandler(*args, **kwargs)
             
-            self._ikeys = self.mapping.getHandle(self._key_node_name)
+            if self.mapping._block_size < 96:
+                raise Exceptions.InvalidFormat(
+                    'Shelve mapping block size must be at least 96 bytes.'
+                )
             
-            if first:
-                self._write_keys(set())
+            self._key_node_name = self._hash_key(b'__KEYS__')
+            
+            with self.mapping.lock:
+                first = False
+                if not self.mapping.nodeExists(self._key_node_name):
+                    self.mapping.makeNode(self._key_node_name)
+                    first = True
+                
+                self._ikeys = self.mapping.getHandle(self._key_node_name)
+                
+                if first:
+                    self._write_keys(set())
         
     def __getitem__(self, key):
-        with self.mapping.lock:
-            key = self._hash_key(pickle.dumps(key))
-            if not self.mapping.nodeExists(key):
-                raise KeyError(key)
-            
-            return pickle.loads(self.mapping.getHandle(key).read())
+        with self._lock:
+            with self.mapping.lock:
+                key = self._hash_key(pickle.dumps(key))
+                if not self.mapping.nodeExists(key):
+                    raise KeyError(key)
+                
+                return pickle.loads(self.mapping.getHandle(key).read())
         
     def _get_keys(self):
         with self.mapping.lock:
             if 'skeys' not in self.mapping.lock.cache:
                 self._ikeys.seek(0)
                 self.mapping.lock.cache['skeys'] = pickle.loads(
-                    lz4.frame.decompress(self._ikeys.read())
+                    zlib.decompress(self._ikeys.read())
                 )
             
             return self.mapping.lock.cache['skeys']
             
     def _write_keys(self, kr):
         with self.mapping.lock:
-            fin = lz4.frame.compress(pickle.dumps(kr))
-            
-            self._ikeys.seek(0)
-            self._ikeys.truncate(len(fin))
-            self._ikeys.write(fin)
-            
             try:
-                del self.mapping.lock.cache['skeys']
+                kro = self.mapping.lock.cache['skeys']
             except KeyError:
-                pass
+                kro = None
+                
+            if kr != kro:
+                fin = zlib.compress(pickle.dumps(kr))
+                
+                self._ikeys.seek(0)
+                self._ikeys.truncate(len(fin))
+                self._ikeys.write(fin)
+                
+                try:
+                    del self.mapping.lock.cache['skeys']
+                except KeyError:
+                    pass
         
     def __delitem__(self, key):
-        with self.mapping.lock:
-            kr = self._get_keys()
-            
-            if key not in kr:
-                raise Exceptions.NodeDoesNotExist('!RMNOD Key: {0}'.format(
-                    key.hex()
-                ))
+        with self._lock:
+            with self.mapping.lock:
+                if self.trackkeys:
+                    kr = self._get_keys()
+                    
+                    if key not in kr:
+                        raise Exceptions.NodeDoesNotExist(
+                            '!RMNOD Key: {0}'.format(
+                                key.hex()
+                            )
+                        )
+                        
+                    kr.remove(key)
+                    self._write_keys(kr)
                 
-            kr.remove(key)
-            self._write_keys(kr)
-            
-            self._ikeys.seek(0)
-            self._ikeys.truncate(len(fin))
-            self._ikeys.write(fin)
-            
-            key = self._hash_key(pickle.dumps(key))
-            self.mapping.removeNode(key)
+                key = self._hash_key(pickle.dumps(key))
+                self.mapping.removeNode(key)
         
     def __setitem__(self, key, value):
-        with self.mapping.lock:
-            kr = self._get_keys()
-            
-            kr.add(key)
-            self._write_keys(kr)
-            
-            key = self._hash_key(pickle.dumps(key))
-            if not self.mapping.nodeExists(key):
-                self.mapping.makeNode(key)
-            
-            handle = self.mapping.getHandle(key)
-            
-            pickval = pickle.dumps(value)
-            handle.truncate(len(pickval))
-            handle.write(pickval)
+        with self._lock:
+            with self.mapping.lock:
+                if self.trackkeys:
+                    kr = self._get_keys()
+                    
+                    kr.add(key)
+                    self._write_keys(kr)
+                
+                key = self._hash_key(pickle.dumps(key))
+                if not self.mapping.nodeExists(key):
+                    self.mapping.makeNode(key)
+                
+                handle = self.mapping.getHandle(key)
+                
+                pickval = pickle.dumps(value)
+                handle.truncate(len(pickval))
+                handle.write(pickval)
         
     def __iter__(self):
-        with self.mapping.lock:
-            return iter(self._get_keys())
+        if self.trackkeys:
+            with self._lock:
+                with self.mapping.lock:
+                    return iter(self._get_keys())
+        else:
+            return None
             
     def __len__(self):
-        with self.mapping.lock:
-            return len(self._get_keys())
+        if self.trackkeys:
+            with self._lock:
+                with self.mapping.lock:
+                    return len(self._get_keys())
+        else:
+            return None
         
     def _hash_key(self, key):
         return hashlib.sha256(key).digest()
@@ -796,16 +984,70 @@ class InstanceLockedShelve(Shelve):
         return self
         
     def __exit__(self, *args, **kwargs):
-        self.mapping.lock.release()
+        self.mapping.lock.release(force = True)
+        
+    def __del__(self, *args, **kwargs):
+        try:
+            self.mapping.lock.release(force = True)
+        except:
+            pass
 
 if __name__ == '__main__':
-    import code, progressbar
+    import progressbar
+    import matplotlib.pyplot as plt
     
-    with InstanceLockedShelve('test.yun') as x:
-        for _ in progressbar.progressbar(range(1024)):
+    FILENAME = '/tmp/test.yun'
+    TEST_VALUES = [os.urandom(2 ** x) for x in range(16)]
+        
+    def pulverise_original():
+        try:
+            os.remove(FILENAME)
+        except:
+            pass
+        
+    def test(x, dotimes):
+        indexes = list(range(dotimes))
+        times = []
+        
+        for _ in progressbar.progressbar(indexes):
             key = os.urandom(8)
-            value = os.urandom(8)
-            x[key] = value
-            assert x[key] == value
+            
+            start = time.perf_counter()
+            operations = 0
+            for value in TEST_VALUES:
+                x[key] = value
+                assert x[key] == value
+                operations += 2
+            end = ((time.perf_counter() - start) / operations) * 1000
+            times.append(end)
+            
+        return indexes, times
+    
+    fig, ax = plt.subplots()
+    
+    pulverise_original()
+    sobj = Shelve(FILENAME)
+    sobj.trackkeys = False
+    ax.plot(*test(sobj, 1024), label='keyless_cooperative')
+    
+    pulverise_original()
+    sobj = Shelve(FILENAME)
+    ax.plot(*test(sobj, 1024), label='cooperative')
+    
+    pulverise_original()
+    with InstanceLockedShelve(FILENAME) as sobj:
+        ax.plot(*test(sobj, 1024), label='lockhogger')
+        
+    pulverise_original()
+    with InstanceLockedShelve(FILENAME) as sobj:
+        sobj.trackkeys = False
+        ax.plot(*test(sobj, 1024), label='keyless_lockhogger')
+    
+    ax.set_xlabel('Created Indexes in Database')
+    ax.set_ylabel('Time to perform a single database operation (milliseconds)')
+    ax.legend()
+    plt.show()
+    
+    pulverise_original()
     
     # code.interact(local=dict(globals(), **locals()))
